package org.openxma.dsl.platform.config;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.LinkedHashSet;
import java.util.Set;

import org.springframework.beans.factory.BeanDefinitionStoreException;
import org.springframework.beans.factory.xml.DefaultBeanDefinitionDocumentReader;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.ResourcePatternUtils;
import org.springframework.util.StringUtils;
import org.w3c.dom.Element;

import at.spardat.properties.XProperties;

/**
 * Extension of the default {@link DefaultBeanDefinitionDocumentReader}
 * implementation to allow the specification of additional query parameter to
 * control the optional import of resources (e.g. ?optional=true) and to enable
 * the resolving of <code>XProperties</code> properties in addition to the
 * exisiting support for system and environment properties/variables.
 */
public class ConfigurableBeanDefinitionDocumentReader extends DefaultBeanDefinitionDocumentReader {
    /** Prefix for system property placeholders: "${" */
    public static final String PLACEHOLDER_PREFIX = "${";
    /** Suffix for system property placeholders: "}" */
    public static final String PLACEHOLDER_SUFFIX = "}";

    @Override
    protected void importBeanDefinitionResource(Element ele) {
        String location = ele.getAttribute(RESOURCE_ATTRIBUTE);
        if (!StringUtils.hasText(location)) {
            getReaderContext().error("Resource location must not be empty", ele);
            return;
        }
        // Resolve placeholders as system or xproperties property
        location = resolvePlaceholders(location);
        boolean optional = isOptional(ele, location);

        if (ResourcePatternUtils.isUrl(location)) {
            try {
                Set actualResources = new LinkedHashSet(4);
                int importCount = getReaderContext().getReader().loadBeanDefinitions(location, actualResources);
                if (logger.isDebugEnabled()) {
                    logger.debug("Imported " + importCount + " bean definitions from URL location [" + location + "]");
                }
                Resource[] actResArray = (Resource[]) actualResources.toArray(new Resource[actualResources.size()]);
                getReaderContext().fireImportProcessed(location, actResArray, extractSource(ele));
            } catch (BeanDefinitionStoreException ex) {
                if (!optional) {
                    getReaderContext().error("Failed to import bean definitions from URL location [" + location + "]",
                            ele, ex);
                }
            }
        } else {
            // No URL -> considering resource location as relative to the
            // current file.
            try {
                Resource relativeResource = getReaderContext().getResource().createRelative(location);
                int importCount = getReaderContext().getReader().loadBeanDefinitions(relativeResource);
                if (logger.isDebugEnabled()) {
                    logger.debug("Imported " + importCount + " bean definitions from relative location [" + location
                            + "]");
                }
                getReaderContext().fireImportProcessed(location, new Resource[] { relativeResource },
                        extractSource(ele));
            } catch (IOException ex) {
                if (!optional) {
                    getReaderContext().error(
                            "Invalid relative resource location [" + location + "] to import bean definitions from",
                            ele, ex);
                }
            } catch (BeanDefinitionStoreException ex) {
                if (!optional) {
                    getReaderContext().error(
                            "Failed to import bean definitions from relative location [" + location + "]", ele, ex);
                }
            }
        }
    }

    /**
     * Resolve ${...} placeholders in the given text, replacing them with
     * corresponding system property values.
     *
     * @param text
     *            the String to resolve
     * @return the resolved String
     * @see #PLACEHOLDER_PREFIX
     * @see #PLACEHOLDER_SUFFIX
     */
    public String resolvePlaceholders(String text) {
        StringBuffer buf = new StringBuffer(text);
        int startIndex = buf.indexOf(PLACEHOLDER_PREFIX);
        while (startIndex != -1) {
            int endIndex = buf.indexOf(PLACEHOLDER_SUFFIX, startIndex + PLACEHOLDER_PREFIX.length());
            if (endIndex != -1) {
                String placeholder = buf.substring(startIndex + PLACEHOLDER_PREFIX.length(), endIndex);
                int nextIndex = endIndex + PLACEHOLDER_SUFFIX.length();
                try {
                    String propVal = System.getProperty(placeholder);
                    if (propVal == null) {
                        // Fall back to searching the system environment.
                        propVal = System.getenv(placeholder);
                    }
                    if (propVal == null) {
                        XProperties.getNode(ConfigurableBeanDefinitionDocumentReader.class);
                        // Fall back to searching XProperties
                        propVal = XProperties.getRoot().get(placeholder);
                    }
                    if (propVal != null) {
                        buf.replace(startIndex, endIndex + PLACEHOLDER_SUFFIX.length(), propVal);
                        nextIndex = startIndex + propVal.length();
                    } else {
                        System.err
                                .println("Could not resolve placeholder '"
                                        + placeholder
                                        + "' in ["
                                        + text
                                        + "] as system property: neither system property, xproperties nor environment variable found");
                    }
                } catch (Throwable ex) {
                    System.err.println("Could not resolve placeholder '" + placeholder + "' in [" + text
                            + "] as system or xproperties property: " + ex);
                }
                startIndex = buf.indexOf(PLACEHOLDER_PREFIX, nextIndex);
            } else {
                startIndex = -1;
            }
        }
        return buf.toString();
    }

    /**
     * Parse the given location as <code>URI</code> to see if the import should
     * be handled optionally.
     *
     * @param ele
     *            the 'import' <code>Element</code>
     * @return true or false wheter the given import should be treated as
     *         optional
     */
    private boolean isOptional(Element ele, String location) {
        try {
            URI uri = new URI(location);
            String query = uri.getQuery();
            return query != null && query.split("=").length == 2 && query.split("=")[0].equals("optional")
                    && Boolean.parseBoolean(query.split("=")[1]);
        } catch (URISyntaxException ex) {
            getReaderContext().warning("Failed to parse URI for location [" + location + "]", ele, ex);
        }
        return false;
    }

}
