package org.openxma.dsl.platform.service;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import org.openxma.dsl.platform.dao.EntityFactory;
import org.openxma.dsl.platform.dao.EntityFactoryRegistry;
import org.openxma.dsl.platform.service.impl.SimpleMapperContext;
import org.springframework.core.NamedInheritableThreadLocal;
import org.springframework.core.NamedThreadLocal;

@SuppressWarnings("unchecked")
public class MapperContextHolder {
    private static final ThreadLocal<MapperContext> mapperContextHolder = new NamedThreadLocal("Mapper context");
    private static final ThreadLocal<MapperContext> inheritableMapperContextHolder = new NamedInheritableThreadLocal(
            "Mapper context");

    /**
     * Reset the MapperContext for the current thread.
     */
    public static void reset() {
        mapperContextHolder.set(null);
        inheritableMapperContextHolder.set(null);
    }

    /**
     * Associate the given MapperContext with the current thread, <i>not</i> exposing it as inheritable for child
     * threads.
     * 
     * @param mapperContext
     *            the current MapperContext, or <code>null</code> to reset the thread-bound context
     */
    public static void setMapperContext(MapperContext mapperContext) {
        setMapperContext(mapperContext, false);
    }

    /**
     * Associate the given MapperContext with the current thread.
     * 
     * @param mapperContext
     *            the current MapperContext, or <code>null</code> to reset the thread-bound context
     * @param inheritable
     *            whether to expose the MapperContext as inheritable for child threads (using an
     *            {@link java.lang.InheritableThreadLocal})
     */
    public static void setMapperContext(MapperContext mapperContext, boolean inheritable) {
        if (inheritable) {
            inheritableMapperContextHolder.set(mapperContext);
            mapperContextHolder.set(null);
        } else {
            mapperContextHolder.set(mapperContext);
            inheritableMapperContextHolder.set(null);
        }
    }

    /**
     * Return the MapperContext associated with the current thread, if any.
     * 
     * @return the current MapperContext, or <code>null</code> if none
     */
    public static MapperContext getMapperContext() {
        MapperContext mapperContext = (MapperContext) mapperContextHolder.get();
        if (mapperContext == null) {
            mapperContext = (MapperContext) inheritableMapperContextHolder.get();
        }
        return mapperContext;
    }

    /**
     * Associate the given EntityFactoryRegistry with the current thread.
     * <p>
     * Will implicitly create a MapperContext for the given EntityFactoryRegistry, <i>not</i> exposing it as inheritable
     * for child threads.
     * 
     * @param registry
     *            the current EntityFactoryRegistry, or <code>null</code> to reset the thread-bound context
     * @see SimpleMapperContext#SimpleMapperContext(java.util.EntityFactoryRegistry)
     */
    public static void setRegistry(EntityFactoryRegistry registry) {
        setRegistryWithContext(registry, new HashMap(), false);
    }

    /**
     * Associate the given EntityFactoryRegistry with the current thread.
     * <p>
     * Will implicitly create a MapperContext for the given EntityFactoryRegistry.
     * 
     * @param registry
     *            the current EntityFactoryRegistry, or <code>null</code> to reset the thread-bound context
     * @param inheritable
     *            whether to expose the MapperContext as inheritable for child threads (using an
     *            {@link java.lang.InheritableThreadLocal})
     * @see SimpleMapperContext#SimpleMapperContext(EntityFactoryRegistry)
     */
    public static void setRegistryWithContext(EntityFactoryRegistry registry, Map context, boolean inheritable) {
        MapperContext mapperContext = (registry != null ? new SimpleMapperContext(registry, context) : null);
        setMapperContext(mapperContext, inheritable);
    }

    /**
     * Return the EntityFactoryRegistry associated with the current thread, if any, or the system default
     * EntityFactoryRegistry else.
     * 
     * @return the current EntityFactoryRegistry, or the system default EntityFactoryRegistry if no specific
     *         EntityFactoryRegistry has been associated with the current thread
     * @see MapperContext#getEntityFactoryRegistry()
     */
    public static EntityFactoryRegistry getEntityFactoryRegistry() {
        MapperContext mapperContext = getMapperContext();
        return (mapperContext != null ? mapperContext.getEntityFactoryRegistry() : null);
    }

    /**
     * @param <T>
     *            the generic type of the entity to be created
     * @param entityClass
     *            the class of the entity to be generated
     * @return an instance of type <code>T</code> which will be created by an instance of
     *         {@link org.openxma.dsl.platform.dao.EntityFactory}
     * 
     */
    public static <T> T createEntity(Class<T> entityClass) {
        return ((EntityFactory<T>) MapperContextHolder.getEntityFactoryRegistry().getEntityFactory(entityClass))
                .createEntity(null);
    }

    public static <T> T createComplexType(Class<T> complexTypeClass) {
        EntityFactory<T> ef = ((EntityFactory<T>) MapperContextHolder.getEntityFactoryRegistry().getEntityFactory(
                complexTypeClass));
        if (null != ef)
            return ef.createEntity(null);

        try {
            return complexTypeClass.getConstructor(new Class[] {}).newInstance(new Object[] {});
        } catch (SecurityException e) {
            throw new RuntimeException("no arg constructor not allowed for class: '" + complexTypeClass.getName()
                    + "' message: '" + e.getMessage() + "'");
        } catch (NoSuchMethodException e) {
            throw new RuntimeException("missing constructor for class: '" + complexTypeClass.getName() + "' message: '"
                    + e.getMessage() + "'");
        } catch (InstantiationException e) {
            throw new RuntimeException("InstantiationException while constructing class: '"
                    + complexTypeClass.getName() + "' message: '" + e.getMessage() + "'");
        } catch (IllegalAccessException e) {
            throw new RuntimeException("IllegalAccessException while constructing class: '"
                    + complexTypeClass.getName() + "' message: '" + e.getMessage() + "'");
        } catch (InvocationTargetException e) {
            throw new RuntimeException("InvocationTargetException while constructing class: '"
                    + complexTypeClass.getName() + "' message: '" + e.getMessage() + "'");
        }
    }

    /**
     * @deprecated use {@link #getContext()} instead
     */
    public static Map getStateMap() {
        return getContext();
    }

    /**
     * Return the MappingContext associated with the current thread, if any, or null else.
     * 
     * @return the current MappingContext, or null
     * @see MapperContext#getStateMap()
     */
    public static Map<Object, Object> getContext() {
        MapperContext mapperContext = getMapperContext();
        return (mapperContext != null ? mapperContext.getContext() : null);
    }

    public static boolean isContext(String contextName) {
        return contextName.equals(MapperContextHolder.getContext().get("context"));
    }

    public static final String MAPPER_OPERATION_PREFIX = "mapper:";
    public static final String CONTEXT = "context";

    /**
     * @return <code>true</code> if the <code>MapperContextHolder</code> holds a value which indicates that the current
     *         mapping operation is triggered by a mapper instead of a dtoAssembler.
     */
    public static boolean isMapperOperation() {
        if (null == MapperContextHolder.getContext() || MapperContextHolder.getContext().size() == 0)
            return false;

        for (Iterator<Entry<Object, Object>> iter = MapperContextHolder.getContext().entrySet().iterator(); iter
                .hasNext();) {
            Entry<Object, Object> next = iter.next();
            if (next.getKey() instanceof String) {
                String key = (String) next.getKey();
                if (key.startsWith(MAPPER_OPERATION_PREFIX))
                    return true;
            }
        }
        return false;
    }

    public static MapperConfig getMapperConfig() {
        if (null == MapperContextHolder.getContext() || MapperContextHolder.getContext().size() == 0)
            return null;

        for (Iterator<Entry<Object, Object>> iter = MapperContextHolder.getContext().entrySet().iterator(); iter
                .hasNext();) {
            Entry<Object, Object> next = iter.next();
            if (next.getKey() instanceof String) {
                String key = (String) next.getKey();
                if (key.startsWith(MAPPER_OPERATION_PREFIX) && next.getValue() instanceof MapperConfig) {
                    return (MapperConfig) next.getValue();
                }
            }
        }
        return null;
    }

    /**
     * @param mapperName
     *            the name of the mapper the context should be created for
     * @param mapperConfig
     *            the config for the mapper to be available for mapping
     * @return a <code>Map</code> which contains as key a {@link String} which looks like {@link MapperContextHolder}
     *         .MAPPER_OPERATION_PREFIX+<code>mapperName</code>
     * @throws {@link IllegalArgumentException} if <code>mapperName</code> is null or empty or consists only of blanks
     */
    public static void setMapperContext(String mapperName, MapperConfig mapperConfig) {
        if (null == mapperName || mapperName.trim().length() == 0)
            throw new IllegalArgumentException("missing mapperName");

        Map<Object,Object> ctx = new HashMap<Object, Object>(1);
        ctx.put(MapperContextHolder.MAPPER_OPERATION_PREFIX + mapperName, mapperConfig);
        MapperContextHolder.setMapperContext(new SimpleMapperContext(null, ctx));
    }
}